/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file DatatypeTrack.cpp
 *       @see MUST::DatatypeTrack.
 *
 *  @date 10.02.2011
 *  @author Tobias Hilbrich, Joachim Protze
 */

#include "GtiMacros.h"

#include "DatatypeTrack.h"

#include <sstream>

using namespace must;

mGET_INSTANCE_FUNCTION(DatatypeTrack);
mFREE_INSTANCE_FUNCTION(DatatypeTrack);
mPNMPI_REGISTRATIONPOINT_FUNCTION(DatatypeTrack);



//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//=============================
// DatatypeTrack
//=============================
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

//=============================
// Constructor
//=============================
DatatypeTrack::DatatypeTrack (const char* instanceName)
: TrackBase<Datatype, MustDatatypeType, MustMpiDatatypePredefined, DatatypeTrack, I_DatatypeTrack> (instanceName),
  myPredefinedInfos (),
  myUbInfo (NULL),
  myLbInfo (NULL),
  CacheHitCount (0),
  CacheMissCount (0)
{
    //Get the GroupTrack module
    if (myFurtherMods.size () < 1)
    {
        std::cerr << "Error: the CommTrack module needs the GroupTrack module as a child, but it was not specified." << std::endl;
        assert (0);
    }

    myBCoMod = (I_BaseConstants*) myFurtherMods[0];
}

//=============================
// Destructor
//=============================
DatatypeTrack::~DatatypeTrack ()
{
    //Notify HandleInfoBase of ongoing shutdown
    HandleInfoBase::disableFreeForwardingAcross();

    //We don't need to free the BaseConstants module, this will be done by TrackBase

#ifdef MUST_DEBUG
    if (CacheHitCount+CacheMissCount>0)
        std::cout << "Cache matches: " << CacheHitCount << " misses: " << CacheMissCount << std::endl;
#endif
}

//=============================
// commit
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::commit (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType type
)
{
    HandleMap::iterator pos = findUserHandle (pId, type);
    //If not a user handle, nothing to do (we do not detect errrors)
    if (pos == myUserHandles.end())
        return GTI_ANALYSIS_SUCCESS;

    //Apply the commit
    pos->second->commit(pId, lId);

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// free
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::free (
        MustParallelId pId,
        MustLocationId lId,
        MustDatatypeType type
)
{
//     HandleMap::iterator pos = findUserHandle (pId, type);
// 
//     //If not a user handle, nothing to do (we do not detect errrors)
//     if (pos == myUserHandles.end())
//         return GTI_ANALYSIS_SUCCESS;
// 
//     //Apply the free
//     pos->second->erase();
//     myUserHandles.erase(pos);
//     myLastQuery = myUserHandles.end();
    removeUserHandle(pId, type);

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeContiguous
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeContiguous (
        MustParallelId pId,
        MustLocationId lId,
        int count,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullContiguousTypeInfo *info = new FullContiguousTypeInfo (this, pId, lId, count, oldType, oldInfos);

    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeHindexed
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeHindexed (
        MustParallelId pId,
        MustLocationId lId,
        int count,
        int* arrayOfBlocklengths,
        MustAddressType* arrayOfDisplacements,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullHIndexedTypeInfo *info = new FullHIndexedTypeInfo (this, pId, lId, count, arrayOfBlocklengths, arrayOfDisplacements, oldType, oldInfos);
//     FullHIndexedTypeInfo* info = new FullHIndexedTypeInfo ();

    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeIndexed
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeIndexed (
        MustParallelId pId,
        MustLocationId lId,
        int count,
        int* arrayOfBlocklengths,
        int* arrayOfDisplacements,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullIndexedTypeInfo *info = new FullIndexedTypeInfo (this, pId, lId, count, arrayOfBlocklengths, arrayOfDisplacements, oldType, oldInfos);
//     FullIndexedTypeInfo *info = new FullIndexedTypeInfo ();


    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeHvector
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeHvector (
        MustParallelId pId,
        MustLocationId lId,
        int count,
        int blocklength,
        MustAddressType stride,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullHVectorTypeInfo *info = new FullHVectorTypeInfo (this, pId, lId, count, blocklength, stride, oldType, oldInfos);

    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeVector
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeVector (
        MustParallelId pId,
        MustLocationId lId,
        int count,
        int blocklength,
        int stride,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullVectorTypeInfo *info = new FullVectorTypeInfo (this, pId, lId, count, blocklength, stride, oldType, oldInfos);

    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeStruct
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeStruct (
        MustParallelId pId,
        MustLocationId lId,
        int count,
        int* arrayOfBlocklengths,
        MustAddressType* arrayOfDisplacements,
        MustDatatypeType* oldTypes,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    if (count < 1)
        /**
         * @todo test count < 1 for other types? Yes
         */
        //TODO should a type with count = 0 not yield a valid handle ? Shouldn't we at least track that?
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos;
    int i;
    for (i = 0; i < count; i++)
    {
        //old type reasonable ?
        oldInfos.push_back((Datatype*)getPersistentDatatype(pId, oldTypes[i]));
        if (oldInfos.back() == NULL)
            return GTI_ANALYSIS_SUCCESS;
    }

    //Handle the type creation
//     FullStructTypeInfo *info = new FullStructTypeInfo ();
     FullStructTypeInfo *info = new FullStructTypeInfo (this, pId, lId, count, arrayOfBlocklengths, arrayOfDisplacements, oldTypes, oldInfos);


    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeIndexedBlock
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeIndexedBlock (
        MustParallelId pId,
        MustLocationId lId,
        int count,
        int blocklength,
        int* arrayOfDisplacements,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullIndexedBlockTypeInfo *info = new FullIndexedBlockTypeInfo (this, pId, lId, count, blocklength, arrayOfDisplacements, oldType, oldInfos);

    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeResized
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeResized (
        MustParallelId pId,
        MustLocationId lId,
        MustAddressType lb,
        MustAddressType extent,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullResizedTypeInfo *info = new FullResizedTypeInfo (this, pId, lId, lb, extent, oldType, oldInfos);

    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeSubarray
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeSubarray (
        MustParallelId pId,
        MustLocationId lId,
        int ndims,
        int* arrayOfSizes,
        int* arrayOfSubsizes,
        int* arrayOfStarts,
        int order,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullSubarrayTypeInfo *info = new FullSubarrayTypeInfo (this, pId, lId, ndims, arrayOfSizes, arrayOfSubsizes, arrayOfStarts, order, oldType, oldInfos);

    submitUserHandle(pId, newType, info);
//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}

//=============================
// typeDarray
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::typeDarray (
        MustParallelId pId,
        MustLocationId lId,
        int size,
        int rank,
        int ndims,
        int* arrayOfGsizes,
        int* arrayOfDistribs,
        int* arrayOfDargs,
        int* arrayOfPsizes,
        int order,
        MustDatatypeType oldType,
        MustDatatypeType newType
)
{
    //new type already known as a handle ?
    //TODO: handle cases where MPI returns the same handle for equal datatypes being constructed by some rank!
    if (isAlreadyKnown (pId, newType))
        return GTI_ANALYSIS_SUCCESS;

    //old type reasonable ?
    std::vector<Datatype *> oldInfos (1,(Datatype*)getPersistentDatatype(pId, oldType));
    if (oldInfos[0] == NULL)
        return GTI_ANALYSIS_SUCCESS; 

    //Handle the type creation
    FullDarrayTypeInfo *info = new FullDarrayTypeInfo (this, pId, lId, size, rank, ndims, arrayOfGsizes, arrayOfDistribs, arrayOfDargs, arrayOfPsizes, order, oldType, oldInfos);

    submitUserHandle(pId, newType, info);

//     //Add to user map
//     myUserHandles.insert(std::make_pair(
//             std::make_pair(pId2Rank(pId), newType),
//             info));
//     myLastQuery = myUserHandles.end();

    return GTI_ANALYSIS_SUCCESS;
}


//=============================
// getPredefinedName
//=============================
std::string DatatypeTrack::getPredefinedName (MustMpiDatatypePredefined predefined)
{
    switch (predefined)
    {
        case MUST_MPI_CHAR: return "MPI_CHAR"; break;
        case MUST_MPI_SHORT: return "MPI_SHORT"; break;
        case MUST_MPI_INT: return "MPI_INT"; break;
        case MUST_MPI_LONG: return "MPI_LONG"; break;
        case MUST_MPI_UNSIGNED_CHAR: return "MPI_UNSIGNED_CHAR"; break;
        case MUST_MPI_UNSIGNED_SHORT: return "MPI_UNSIGNED_SHORT"; break;
        case MUST_MPI_UNSIGNED: return "MPI_UNSIGNED"; break;
        case MUST_MPI_UNSIGNED_LONG: return "MPI_UNSIGNED_LONG"; break;
        case MUST_MPI_FLOAT: return "MPI_FLOAT"; break;
        case MUST_MPI_DOUBLE: return "MPI_DOUBLE"; break;
        case MUST_MPI_LONG_DOUBLE: return "MPI_LONG_DOUBLE"; break;
        case MUST_MPI_BYTE: return "MPI_BYTE"; break;
        case MUST_MPI_PACKED: return "MPI_PACKED"; break;
        case MUST_MPI_INTEGER: return "MPI_INTEGER"; break;
        case MUST_MPI_REAL: return "MPI_REAL"; break;
        case MUST_MPI_DOUBLE_PRECISION: return "MPI_DOUBLE_PRECISION"; break;
        case MUST_MPI_COMPLEX: return "MPI_COMPLEX"; break;
        case MUST_MPI_LOGICAL: return "MPI_LOGICAL"; break;
        case MUST_MPI_CHARACTER: return "MPI_CHARACTER"; break;
        case MUST_MPI_FLOAT_INT: return "MPI_FLOAT_INT"; break;
        case MUST_MPI_DOUBLE_INT: return "MPI_DOUBLE_INT"; break;
        case MUST_MPI_LONG_INT: return "MPI_LONG_INT"; break;
        case MUST_MPI_2INT: return "MPI_2INT"; break;
        case MUST_MPI_SHORT_INT: return "MPI_SHORT_INT"; break;
        case MUST_MPI_LONG_DOUBLE_INT: return "MPI_LONG_DOUBLE_INT"; break;
        case MUST_MPI_2REAL: return "MPI_2REAL"; break;
        case MUST_MPI_2DOUBLE_PRECISION: return "MPI_2DOUBLE_PRECISION"; break;
        case MUST_MPI_2INTEGER: return "MPI_2INTEGER"; break;
        case MUST_MPI_2COMPLEX: return "MPI_2COMPLEX"; break;
        case MUST_MPI_2DOUBLE_COMPLEX: return "MPI_2DOUBLE_COMPLEX"; break;
        case MUST_MPI_LONG_LONG_INT: return "MPI_LONG_LONG_INT"; break;
        case MUST_MPI_LONG_LONG: return "MPI_LONG_LONG"; break;
        case MUST_MPI_UNSIGNED_LONG_LONG: return "MPI_UNSIGNED_LONG_LONG"; break;
        case MUST_MPI_WCHAR: return "MPI_WCHAR"; break;
        case MUST_MPI_SIGNED_CHAR: return "MPI_SIGNED_CHAR"; break;
        case MUST_MPI_INTEGER1: return "MPI_INTEGER1"; break;
        case MUST_MPI_INTEGER2: return "MPI_INTEGER2"; break;
        case MUST_MPI_INTEGER4: return "MPI_INTEGER4"; break;
        case MUST_MPI_INTEGER8: return "MPI_INTEGER8"; break;
        case MUST_MPI_INTEGER16: return "MPI_INTEGER16"; break;
        case MUST_MPI_REAL2: return "MPI_REAL2"; break;
        case MUST_MPI_REAL4: return "MPI_REAL4"; break;
        case MUST_MPI_REAL8: return "MPI_REAL8"; break;
        case MUST_MPI_REAL16: return "MPI_REAL16"; break;
        case MUST_MPI_DOUBLE_COMPLEX: return "MPI_DOUBLE_COMPLEX"; break;
        case MUST_MPI_COMPLEX8: return "MPI_COMPLEX8"; break;
        case MUST_MPI_COMPLEX16: return "MPI_COMPLEX16"; break;
        case MUST_MPI_COMPLEX32: return "MPI_COMPLEX32"; break;
        case MUST_MPI_LOGICAL1: return "MPI_LOGICAL1"; break;
        case MUST_MPI_LOGICAL2: return "MPI_LOGICAL2"; break;
        case MUST_MPI_LOGICAL4: return "MPI_LOGICAL4"; break;
        case MUST_MPI_LOGICAL8: return "MPI_LOGICAL8"; break;
        case MUST_MPI_LOGICAL16: return "MPI_LOGICAL16"; break;
        case MUST_MPI_UB: return "MPI_UB"; break;
        case MUST_MPI_LB: return "MPI_LB"; break;
    }

    return "Unknown Datatype";
}


//=============================
// fillPredefinedInfos
//=============================
// template < typename FULL_INFO, typename HANDLE_TYPE, typename PREDEFINED_ENUM, class SUPER, class INTERFACE >
// TrackBase<Datatype, MustDatatypeType, MustMpiDatatypePredefined, DatatypeTrack, I_DatatypeTrack> (instanceName),
Datatype* DatatypeTrack::createPredefinedInfo (int predef, MustDatatypeType value)
{
    bool explicitLb=false;
    bool explicitUb=false;
    bool null = false;
    bool predefined = true;
    bool commited = false;
    bool optional = false;
    bool forReduction = false;
    bool boundMarker = false;
    bool c = false;
    bool fortran = false;
    const char* name;

    if (value == myNullValue){
        // MPI_DATATYPE_NULL
        null = true;
        c = true;
        fortran = false;
        optional = false;
        forReduction = false;
        boundMarker = false;
    }
    else
    {
        switch (predef)
        {
            //Elementary C
            case MUST_MPI_CHAR:
            case MUST_MPI_SHORT:
            case MUST_MPI_INT:
            case MUST_MPI_LONG:
            case MUST_MPI_UNSIGNED_CHAR:
            case MUST_MPI_UNSIGNED_SHORT:
            case MUST_MPI_UNSIGNED:
            case MUST_MPI_UNSIGNED_LONG:
            case MUST_MPI_FLOAT:
            case MUST_MPI_DOUBLE:
            case MUST_MPI_LONG_DOUBLE:
                c = true;
                fortran = false;
                optional = false;
                forReduction = false;
                boundMarker = false;
                break;

                //Elementary C & Fortran
            case MUST_MPI_BYTE:
            case MUST_MPI_PACKED:
                c = true;
                fortran = true;
                optional = false;
                forReduction = false;
                boundMarker = false;
                break;

                //Elementary Fortran
            case MUST_MPI_INTEGER:
            case MUST_MPI_REAL:
            case MUST_MPI_DOUBLE_PRECISION:
            case MUST_MPI_COMPLEX:
            case MUST_MPI_LOGICAL:
            case MUST_MPI_CHARACTER:
                c = false;
                fortran = true;
                optional = false;
                forReduction = false;
                boundMarker = false;
                break;

                //Reduction types C
            case MUST_MPI_FLOAT_INT:
            case MUST_MPI_DOUBLE_INT:
            case MUST_MPI_LONG_INT:
            case MUST_MPI_2INT:
            case MUST_MPI_SHORT_INT:
            case MUST_MPI_LONG_DOUBLE_INT:
                c = true;
                fortran = false;
                optional = false;
                forReduction = true;
                boundMarker = false;
                break;

                //Reduction types Fortran (first three non-optional, last two optional)
            case MUST_MPI_2REAL:
            case MUST_MPI_2DOUBLE_PRECISION:
            case MUST_MPI_2INTEGER:
                c = false;
                fortran = true;
                optional = false;
                forReduction = true;
                boundMarker = false;
                break;
            case MUST_MPI_2COMPLEX:
            case MUST_MPI_2DOUBLE_COMPLEX:
                c = false;
                fortran = true;
                optional = true;
                forReduction = true;
                boundMarker = false;
                break;

                //Optional C
            case MUST_MPI_LONG_LONG_INT:
            case MUST_MPI_LONG_LONG:
            case MUST_MPI_UNSIGNED_LONG_LONG:
            case MUST_MPI_WCHAR:
            case MUST_MPI_SIGNED_CHAR:
                c = true;
                fortran = false;
                optional = true;
                forReduction = false;
                boundMarker = false;
                break;

                //Optional Fortran
            case MUST_MPI_INTEGER1:
            case MUST_MPI_INTEGER2:
            case MUST_MPI_INTEGER4:
            case MUST_MPI_INTEGER8:
            case MUST_MPI_INTEGER16:
            case MUST_MPI_REAL2:
            case MUST_MPI_REAL4:
            case MUST_MPI_REAL8:
            case MUST_MPI_REAL16:
            case MUST_MPI_DOUBLE_COMPLEX:
            case MUST_MPI_COMPLEX8:
            case MUST_MPI_COMPLEX16:
            case MUST_MPI_COMPLEX32:
            case MUST_MPI_LOGICAL1:
            case MUST_MPI_LOGICAL2:
            case MUST_MPI_LOGICAL4:
            case MUST_MPI_LOGICAL8:
            case MUST_MPI_LOGICAL16:
                c = false;
                fortran = true;
                optional = true;
                forReduction = false;
                boundMarker = false;
                break;

                //Bound markers
            case MUST_MPI_UB:
                c = true;
                fortran = true;
                optional = false;
                forReduction = false;
                boundMarker = true;
                explicitUb=true;
                break;
            case MUST_MPI_LB:
                c = true;
                fortran = true;
                optional = false;
                forReduction = false;
                boundMarker = true;
                explicitLb=true;
                break;
            default:
                    std::cerr << "Controll completness of base types in " << __FILE__ << "@" << __LINE__ << " missing predef has number " << predef << std::endl;
                    assert (0);

        }
    }

    FullBaseTypeInfo* info = new FullBaseTypeInfo (this, optional, forReduction, boundMarker, null, c, fortran, explicitLb, explicitUb, (MustMpiDatatypePredefined)predef, getPredefinedName((MustMpiDatatypePredefined)predef).c_str());

    if (explicitLb==true)
            myLbInfo = info;
    if (explicitUb==true)
            myUbInfo = info;
    return info;
    ////DEBUG output
    //PredefinedMap::iterator iter2;
    //for (iter2 = myPredefineds.begin(); iter2 != myPredefineds.end(); iter2++)
    //{
    //    std::cout << "Predef: " << getPredefinedName (iter2->second)  << " value " << iter2->first << std::endl;
    //}
}


//=============================
// addPredefinedTypes
//=============================
GTI_ANALYSIS_RETURN DatatypeTrack::addPredefinedTypes (
            		MustDatatypeType datatypeNull,
            		int numPredefs,
            		int* predefinedIds,
            		MustDatatypeType* predefinedValues,
            		MustAddressType *extents,
            		int *alignments)
{
	//== 1) Trigger TrackBase to create the handle to enum mapping
	// This will triger fillPredefinedInfos, which will create the full info
	// Structs for all base types.
	TrackBase<Datatype, MustDatatypeType, MustMpiDatatypePredefined, DatatypeTrack, I_DatatypeTrack>::addPredefineds (
		datatypeNull,
    		numPredefs,
    		predefinedIds,
    		predefinedValues
    		);

	//== 2) Update the extents and alignments according to the given values
// 	std::map <MustDatatypeType, Datatype*>::iterator iter;
	PredefinedInfos::iterator iter;
	for (int i = 0; i < numPredefs; i++)
	{
		iter = myPredefineds.find(predefinedValues[i]);

		if (iter == myPredefineds.end())
		{
#ifdef MUST_DEBUG
			std::cerr<<"Internal warning in " << __FILE__ << "@" << __LINE__ << " for mpi type (must enum) " << predefinedIds[i] << ", there is no full info for this type, likely due to two MPI datatypes havin the same handle (e.g. MPI_LONG_LONG and MPI_LONG_LONG_INT in OpenMPI)."<< std::endl;
#endif
			continue;
		}

		((FullBaseTypeInfo*)iter->second)->setSizes(extents[i], alignments[i]);
// 		std::cout << "DT Info, " << iter->second->getPredefinedName() << " extent=" << iter->second->isNull() << " alignment=" << iter->second->getAlignment() << std::endl;
	}

	return GTI_ANALYSIS_SUCCESS;
}


//=============================
// getUbInfo
//=============================
FullBaseTypeInfo* DatatypeTrack::getUbInfo (void)
{
	return myUbInfo;
}

//=============================
// getLbInfo
//=============================
FullBaseTypeInfo* DatatypeTrack::getLbInfo (void)
{
	return myLbInfo;
}

//=============================
// getBCoMod
//=============================
I_BaseConstants* DatatypeTrack::getBCoMod()
{
    return myBCoMod;
}

//=============================
// getDatatype
//=============================
I_Datatype * DatatypeTrack::getDatatype (
        MustParallelId pId,
        MustDatatypeType type)
{
    return getHandleInfo(pId, type);
}

//=============================
// getDatatype
//=============================
I_Datatype * DatatypeTrack::getDatatype (
        int rank,
        MustDatatypeType type)
{
    return getHandleInfo(rank, type);
}

//=============================
// getPersistentDatatype
//=============================
I_DatatypePersistent * DatatypeTrack::getPersistentDatatype (
        MustParallelId pId,
        MustDatatypeType type)
{
    Datatype* ret = getHandleInfo(pId, type);

    if (ret)
        ret->incRefCount();
    return ret;
}

//=============================
// getPersistentDatatype
//=============================
I_DatatypePersistent * DatatypeTrack::getPersistentDatatype (
        int rank,
        MustDatatypeType type)
{
    Datatype* ret = getHandleInfo(rank, type);
    ret->incRefCount();
    return ret;
}

//=============================
// getRemoteDatatype
//=============================
I_Datatype* DatatypeTrack::getRemoteDatatype (
        MustParallelId pId,
        MustRemoteIdType remoteId)
{
    return getRemoteDatatype (pId2Rank(pId), remoteId);
}

//=============================
// getRemoteDatatype
//=============================
I_Datatype* DatatypeTrack::getRemoteDatatype (
        int rank,
        MustRemoteIdType remoteId)
{
    Datatype* ret = getRemoteIdInfo(rank,remoteId);
    return ret;
}

//=============================
// getPersistentRemoteDatatype
//=============================
I_DatatypePersistent* DatatypeTrack::getPersistentRemoteDatatype (
        MustParallelId pId,
        MustRemoteIdType remoteId)
{
    return getPersistentRemoteDatatype (pId2Rank(pId), remoteId);
}

//=============================
// getPersistentRemoteDatatype
//=============================
I_DatatypePersistent* DatatypeTrack::getPersistentRemoteDatatype (
        int rank,
        MustRemoteIdType remoteId)
{
    Datatype* ret = getRemoteIdInfo(rank,remoteId);
    if (ret) ret->copy();
    return ret;
}

//=============================
// getPersistentRemoteDatatype
//=============================
bool DatatypeTrack::passDatatypeAcross (
        MustParallelId pId,
        MustDatatypeType datatype,
        int toPlaceId)
{
    return passDatatypeAcross (pId2Rank(pId), datatype, toPlaceId);
}

//=============================
// getPersistentRemoteDatatype
//=============================
bool DatatypeTrack::passDatatypeAcross (
        int rank,
        MustDatatypeType datatype,
        int toPlaceId)
{
    Datatype* info = getHandleInfo(rank, datatype);

    if (!info)
        return false; //Invalid call, type not known

    return passDatatypeAcrossInternal (rank, info, toPlaceId, NULL, true, datatype);
}

//=============================
// getPersistentRemoteDatatype
//=============================
bool DatatypeTrack::passDatatypeAcross (
        int rank,
        I_Datatype* datatype,
        int toPlaceId,
        MustRemoteIdType *pOutRemoteId)
{
    //Valid info?
    if (!datatype)
        return false;

    //Cast to internal representation
    Datatype* info = (Datatype*) datatype;

    //Do we have a handle for this still?
    MustDatatypeType handle = 0;
    bool hasHandle = getHandleForInfo (rank, info, &handle);

    return passDatatypeAcrossInternal (rank, info, toPlaceId, pOutRemoteId, hasHandle, handle);
}

//=============================
// passDatatypeAcrossInternal
//=============================
bool DatatypeTrack::passDatatypeAcrossInternal (
        int rank,
        Datatype* datatype,
        int toPlaceId,
        MustRemoteIdType *pOutRemoteId,
        bool hasHandle,
        MustDatatypeType handle)
{
    //TODO see CommTrack::passCommAcrossInternal for implementation hints
    return true;
}

/*EOF*/
