/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file StridedBlock.h
 *       @see StridedBlock.
 *
 *  @date 17.08.2011
 *  @author Joachim Protze
 */

#include <set>
#include <stdint.h>
#include "MustTypes.h"

#ifndef STRIDEDBLOCKS_H
#define STRIDEDBLOCKS_H

namespace must
{
    class I_Datatype;


    struct StridedBlock
    {
        MustAddressType first; // begin
        MustAddressType second; // end
        MustAddressType pos; // byte position within the serialized datatype
        bool merged; // means the block entry was merged non contiguous (either overlapping or non-chronological)
        int repetition;
        int count;
        MustAddressType blocksize;
        MustAddressType stride;
        bool operator<(const StridedBlock& r) const;
        
        // constructors
        // copyconst
        StridedBlock(const StridedBlock& o) : first(o.first), second(o.second), pos(o.pos), merged(o.merged), repetition(o.repetition), count(o.count), blocksize(o.blocksize), stride(o.stride){}
        
        StridedBlock(const MustAddressType& l) : first(0), second(l), pos(0), merged(false), repetition(0), count(1), blocksize(l), stride(0) {}

        StridedBlock(const MustAddressType& b, const MustAddressType& p, const bool& m, const int& r, const int& c, const MustAddressType& block, const MustAddressType& stride) : first(b), second(b+(c-1)*stride+block), pos(p), merged(m), repetition(r), count(c), blocksize(block), stride(stride){}
        
        StridedBlock(const StridedBlock& o, const MustAddressType& offset, const int& rep) : first(o.first+offset), second(o.second+offset), pos(o.pos), merged(o.merged), repetition(rep), count(o.count), blocksize(o.blocksize), stride(o.stride){}

        StridedBlock(const StridedBlock& o, const MustAddressType& offset, const MustAddressType& posoffset, const int& rep) : first(o.first+offset), second(o.second+offset), pos(o.pos+posoffset), merged(o.merged), repetition(rep), count(o.count), blocksize(o.blocksize), stride(o.stride){}
        
        /* StridedBlock(const MustAddressType& b, const MustAddressType& e, const MustAddressType& p, const int& c, const MustAddressType& block, const MustAddressType& stride) : first(b), second(e), pos(p), merged(false), repetition(0), count(c), blocksize(block), stride(stride) {}

        StridedBlock(const MustAddressType& b, const MustAddressType& e, const MustAddressType& p, const bool& m, const int& c, const MustAddressType& block, const MustAddressType& stride) : first(b), second(e), pos(p), merged(m), repetition(0), count(c), blocksize(block), stride(stride) {}
        
        StridedBlock(const StridedBlock& o, const MustAddressType& b, const MustAddressType& e, const MustAddressType& p, const int& r) : first(b), second(e), pos(p), merged(o.merged), repetition(r), count(o.count), blocksize(o.blocksize), stride(o.stride){}*/
        bool overlapPos(MustAddressType b, MustAddressType e, int * posB) const;
        
        bool overlaps(MustAddressType b, MustAddressType e) const;
        
        bool overlapPos(const StridedBlock& o, int& posA, int& posB) const;
        bool overlaps(const StridedBlock & r) const;
    };
    
    struct memInterval : StridedBlock
    {
        MustRequestType request;
        bool isSend;
        I_Datatype * type;
        
        // copyconst
        memInterval(const memInterval& o) : StridedBlock(o.first, o.pos, o.merged, o.repetition, o.count, o.blocksize, o.stride), request(o.request), isSend(o.isSend), type(o.type) {}
        
        memInterval(const StridedBlock& sBlock, const MustAddressType& offset, const MustRequestType& r, const bool& s, I_Datatype *const &t, const int& rep) : StridedBlock(sBlock, offset, rep), request(r), isSend(s), type(t) {}
        
        memInterval(const StridedBlock& sBlock, const MustAddressType& offset, const MustAddressType& posoffset, const MustRequestType& r, const bool& s, I_Datatype *const &t, const int& rep) : StridedBlock(sBlock, offset, posoffset, rep), request(r), isSend(s), type(t) {}
        
    };
    
    typedef std::multiset<memInterval> MustMemIntervalListType;


    /**
    * Datatype used for Blocklists.
    */
    typedef std::multiset<StridedBlock> MustStridedBlocklistType;

    /**
    * create a blocklist by replication
    */
    MustStridedBlocklistType buildStridedBlocklist(const MustStridedBlocklistType& blocklist, MustAddressType extent, MustAddressType size, MustAddressType offset, MustAddressType posOffset, MustAddressType blocklength, MustAddressType stride, MustAddressType count);

    MustMemIntervalListType buildMemIntervallist(const MustStridedBlocklistType& blocklist, const MustAddressType& extent, const MustAddressType& size, const MustAddressType& offset, const MustRequestType& request, const bool& isSend, I_Datatype *const& type, const int& count);

    /**
    * Blocklist Info package for return values.
    */
    struct BlockInfo
    {
        MustStridedBlocklistType stridedBlocklist;
        bool overlapped;
    };
    
    // functions to check lists for overlap
    
    
    bool isOverlapped(const MustStridedBlocklistType& blocklist, MustStridedBlocklistType::iterator& iter, MustStridedBlocklistType::iterator& nextIter);
    
    bool isOverlapped(const MustMemIntervalListType& blocklist, MustMemIntervalListType::iterator& iter, MustMemIntervalListType::iterator& nextIter, const bool& ignoreSend);

    bool isOverlapped(const MustMemIntervalListType& blocklist, MustMemIntervalListType::iterator& iter, MustMemIntervalListType::iterator& nextIter, int& posA, int& posB, const bool& ignoreSend);

    bool isOverlapped(const MustMemIntervalListType& listA, const MustMemIntervalListType& listB, MustMemIntervalListType::iterator& iter, MustMemIntervalListType::iterator& nextIter, const bool& ignoreSend);

    bool isOverlapped(const MustMemIntervalListType& listA, const MustMemIntervalListType& listB, MustMemIntervalListType::iterator& iter, MustMemIntervalListType::iterator& nextIter, int& posA, int& posB, const bool& ignoreSend);

    
    
} // namespace must
#endif /*STRIDEDBLOCKS_H*/


